# frozen_string_literal: true

#
# Author:: Jonathan Hartman (<jonathan.hartman@socrata.com>)
#
# Copyright (C) 2018, Socrata, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

require 'fileutils'
require 'kitchen/loader/yaml'
require 'kitchen/util'
require 'vendor/hash_recursive_merge'
require 'yaml'
require_relative '../microwave/driver'
require_relative '../microwave/platforms'
require_relative '../microwave/provisioner'
require_relative '../microwave/suites'
require_relative '../microwave/transport'
require_relative '../microwave/verifier'

module Kitchen
  module Loader
    # A customized version of Test Kitchen's YAML loader to allow
    # auto-generating a config.
    #
    # @author Jonathan Hartman <jonathan.hartman@socrata.com>
    class Microwave < YAML
      #
      # Write out the generated hash for this loader after instantiating it.
      #
      # (see Kitchen::Loader::YAML.initialize)
      #
      def initialize(options = {})
        super
        write!
      end

      #
      # Don't bother raising an error if the .kitchen.yml doesn't exist.
      #
      # (see Kitchen::Loader::YAML#read)
      #
      def read
        Util.symbolized_hash(combined_hash)
      end

      #
      # Write the merged config out to a file for easier debugging.
      #
      def write!
        path = File.expand_path('.kitchen/.microwave.yml',
                                File.dirname(@config_file))
        FileUtils.mkdir_p(File.dirname(path))
        File.write(path, Util.stringified_hash(read).to_yaml)
      end

      #
      # Add a fourth layer of config at the top and then merge the regular
      # three layers of .kitchen.ymls into it. Then delete any excluded
      # platforms that have been set globally via:
      #
      # ---
      # microwave:
      #   excludes:
      #     - platform: ubuntu
      #       version: 14.04
      #       chef: 14
      #     - platform: debian
      #       # version: implicit "all"
      #       # chef: implicit "all"
      #
      # (see Kitchen::Loader::YAML#combined_hash)
      #
      def combined_hash
        hsh = microwave_hash.rmerge(super)
        @excludes = hsh['microwave']['excludes']
        hsh
      end

      #
      # Return the new top layer we're going to merge all the traditional YAML
      # layers into. The new layer should have:
      #
      # * An empty set of platforms to be excluded
      # * A static, simple driver config for dokken
      # * A static, simple transport config for dokken
      # * A static, simple provisioner config for dokken
      # * A static, simple verifier config for dokken
      # * A set of platforms generated from metadata.rb's supported
      #   platforms*Chef versions
      # * A set of suites generated from recipes present in the `test` wrapper
      #   cookbook
      #
      # @return [Hash] the top-level Microwave hash for the config
      def microwave_hash
        path = File.dirname(@config_file)
        normalize(
          'microwave' => { 'excludes' => [] },
          'driver' => Kitchen::Microwave::Driver.new,
          'transport' => Kitchen::Microwave::Transport.new,
          'provisioner' => Kitchen::Microwave::Provisioner.new,
          'verifier' => Kitchen::Microwave::Verifier.new,
          'platforms' => Kitchen::Microwave::Platforms.new(path, @excludes),
          'suites' => Kitchen::Microwave::Suites.new(path)
        )
      end
    end
  end
end
